# -*- coding: utf-8 -*-
from __future__ import division, absolute_import
from __future__ import print_function

import logging
import numpy as np
import pandas as pd

from pyscal.constants import SWINTEGERS

from .wateroil import WaterOil
from .gasoil import GasOil


class WaterOilGas(object):

    """A representation of three-phase properties for oil-water-gas

    Use one object for each satnum.

    One WaterOil and one GasOil object will be created, with
    compatible saturation ranges. Access the class members 'wateroil'
    and 'gasoil' directly to add curves.

    All arguments can be defaulted, and all are zero except h.

    Args:
        swirr (float): Irreducible water saturation for capillary pressure
        swl (float): First water saturation point in outputted tables.
        swcr (float): Critical water saturation, water is immobile below this
        sorw (float): Residual oil saturation
        sgcr (float): Critical gas saturation, gas is immobile below this
        h (float): Saturation intervals in generated tables.
        tag (str): Optional text that will be included as comments.
   """

    def __init__(
        self, swirr=0, swl=0.0, swcr=0.0, sorw=0.00, sorg=0, sgcr=0, h=0.01, tag=""
    ):
        """Sets up the saturation range for three phases"""
        self.wateroil = WaterOil(
            swirr=swirr, swl=swl, swcr=swcr, sorw=sorw, h=h, tag=tag
        )
        self.gasoil = GasOil(swirr=swirr, sgcr=sgcr, sorg=sorg, swl=swl, h=h, tag=tag)

    def selfcheck(self):
        """Run selfcheck on both wateroil and gasoil.

        Returns true only if both passes"""
        return self.wateroil.selfcheck() and self.gasoil.selfcheck()

    def SWOF(self, header=True, dataincommentrow=True):
        """Return a SWOF string. Delegated to the wateroil object"""
        return self.wateroil.SWOF(header, dataincommentrow)

    def SGOF(self, header=True, dataincommentrow=True):
        """Return a SGOF string. Delegated to the gasoil object."""
        return self.gasoil.SGOF(header, dataincommentrow)

    def SLGOF(self, header=True, dataincommentrow=True):
        """Return a SLGOF string. Delegated to the gasoil object."""
        return self.gasoil.SLGOF(header, dataincommentrow)

    def SGFN(self, header=True, dataincommentrow=True):
        """Return a SGFN string. Delegated to the gasoil object."""
        return self.gasoil.SGFN(header, dataincommentrow)

    def SWFN(self, header=True, dataincommentrow=True):
        """Return a SWFN string. Delegated to the wateroil object."""
        return self.wateroil.SWFN(header, dataincommentrow)

    def SOF3(self, header=True, dataincommentrow=True):
        """Return a SOF3 string, combining data from the wateroil and
        gasoil objects.

        So - the oil saturation ranges from 0 to 1-swl. The saturation points
        from the WaterOil object is used to generate these
        """
        # Copy of the wateroil data:
        table = pd.DataFrame(self.wateroil.table[["sw", "krow"]])
        table["so"] = 1 - table["sw"]

        # Copy of the gasoil data:
        gastable = pd.DataFrame(self.gasoil.table[["sg", "krog"]])
        gastable["so"] = 1 - gastable["sg"] - self.wateroil.swl

        # Merge WaterOil and GasOil on oil saturation, interpolate for
        # missing data (potentially different sg- and sw-grids)
        sof3table = (
            pd.concat([table, gastable], sort=True)
            .set_index("so")
            .sort_index()
            .interpolate(method="slinear")
            .fillna(method="ffill")
            .fillna(method="bfill")
            .reset_index()
        )
        sof3table["soint"] = list(
            map(int, list(map(round, sof3table["so"] * SWINTEGERS)))
        )
        sof3table.drop_duplicates("soint", inplace=True)

        # The 'so' column has been calculated from floating point numbers
        # and the zero value easily becomes a negative zero, circumvent this:
        zerorow = np.isclose(sof3table["so"], 0.0)
        sof3table.loc[zerorow, "so"] = abs(sof3table.loc[zerorow, "so"])

        string = ""
        if header:
            string += "SOF3\n"
        string += "-- " + self.wateroil.tag + "\n"
        string += "-- " + self.gasoil.tag + "\n"
        string += "-- So Krow Krog\n"
        if dataincommentrow:
            string += self.wateroil.swcomment
            string += self.gasoil.sgcomment

        string += sof3table[["so", "krow", "krog"]].to_csv(
            sep=" ", float_format="%1.5f", header=None, index=False
        )
        string += "/\n"
        return string

    def threephaseconsistency(self):
        """Perform consistency checks on produced curves, similar
        to what Eclipse does at startup

        Returns empty string if no errors catched. Alternatively
        an error description string is returned.

        Possible variation of this function would be
        to throw Exceptions.
        """

        # Eclipse errors:

        # 1: Error in saturation table number 1 at the maximum oil
        # saturation (0.9) krow and krog should both be equal the oil
        # relative permeability for a system with oil and connate
        # water only - but in this case they are different (krow=1.0
        # and krog=0.93)

        errors = ""
        if not np.isclose(
            self.wateroil.table["krow"].max(), self.gasoil.table["krog"].max()
        ):
            errors += "Error: max(krow) is not equal to max(krog)\n"

        # 2: Inconsistent end points in saturation table 1 the maximum
        # gas saturation (0.91) plus the connate water saturation
        # (0.10) must not exceed 1.0
        if self.gasoil.table["sg"].max() + self.wateroil.table["sw"].min() > 1.0:
            errors += "Error: Max(sg) + Swl > 1.0\n"

        # 3: Warning: Consistency problem for gas phase endpoint (krgr > krg)
        # in grid cell (i, j, k) for saturation end-points krgr=1.0
        # krg = 0.49.

        # 4: Warning: Consistency problem for oil phase endpoint (sgu > 1-swl)
        # in grid cell (i, j, k) for saturation endpoints sgu=0.78,
        # swl=0.45, (1-swl) = 0.55

        if len(errors):
            return errors
        else:
            return ""

    def run_eclipse_test(self):
        """Start the Eclipse simulator on a minimal deck in order to
        test the properties of the current WaterOilGas deck"""
        import tempfile
        import os
        import subprocess

        ecldeckbeforeprops = """RUNSPEC
DIMENS
  1 1 1 /
OIL
WATER
GAS
START
  1 'JAN' 2100 /
TABDIM
   2* 10000 /
GRID
DX
   10 /
DY
   10 /
DZ
   50 /
TOPS
   1000 /
PORO
   0.3 /
PERMX
   100 /
PERMY
   100 /
PERMZ
   100 /

PROPS
"""
        ecldeckafterprops = """
DENSITY
  800 1000 1.2 /
PVTW
  1 1 0.0001 0.2 0.00001 /
PVDO
   100 1   1
   150 0.9 1 /
PVDG
   100 1 1
   150 0.9 1 /
ROCK
  100 0.0001 /
SOLUTION
EQUIL
   1000    100     1040    0   1010      0 /"""

        # Generate the finished Eclipse deck.
        ecldeck = ecldeckbeforeprops + self.SWOF() + self.SGOF() + ecldeckafterprops

        tmpdir = tempfile.mkdtemp()
        eclfile = os.path.join(tmpdir, "RELPERMTEST.DATA")
        with open(eclfile, "w") as eclfileh:
            eclfileh.write(ecldeck)
        ecloutput = subprocess.check_output(
            ["/project/res/bin/runeclipse", "-i", eclfile]
        )
        ecloutputlines = ecloutput.split("\n")
        logging.error([x for x in ecloutputlines if "Error" in x or "ERROR" in x])



# -*- coding: utf-8 -*-
"""A module for creating relative permeability input curves for
Eclipse and Nexus.
"""

from __future__ import division, absolute_import
from __future__ import print_function

from .wateroil import WaterOil
from .wateroilgas import WaterOilGas
from .gasoil import GasOil
from .scalrecommendation import SCALrecommendation

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# -*- coding: utf-8 -*-

import copy
import logging
import numpy as np
import pandas as pd

from pyscal.constants import SWINTEGERS

from pyscal import WaterOilGas, WaterOil, GasOil


class SCALrecommendation(object):
    """A SCAL recommendation consists of three OilWaterGas objects,
    tagged low, base and high.

    For a continuum to exist within a SCAL recommendation, relperm
    curves are interpolated pointwise in relperm-vs-saturation
    space. Because the individual low, base and high curves might
    cross, interpolation cannot be performed in the parameter space
    (L, E, T, endpoints).

    If endpoints do not vary, and T has opposite direction than L and
    E going from low to h igh, interpolation in L, E and T-space is
    possible without interpolated curves going outside the area
    contained by low and high curves.

    """

    @staticmethod
    def defaultshandling(key, value, dicts):
        """Helper function for __init__ to fill out missing values in
        dicts with relperm parameter"""
        for dic in dicts:
            if key not in dic:
                dic[key] = value

    def __init__(self, low, base, high, tag, h=0.01):
        """Set up a SCAL recommendation curve set

        You can choose to provide ready-made WaterOilGas objects, or
        dictionaries with LET_properties for initialization.

        If you have LET-properties in a dictionary, the low, base and
        high objects must be of type dict must containing the
        properties for each curve set, in the following keys:

           Lw, Ew, Tw, Lo, Eo, To,
           Lg, Eg, Tg, Log, Eog, Tog,
           krwr
           swirr, swl, sorw, sorg, sgcr

        For oil-water only, you may omit the LET parameters for gas and oil-gas

        If low, base and high are of type WaterOilGas objects, these
        will be used as is.
        """

        self.h = h
        self.tag = tag

        if isinstance(low, dict) and isinstance(base, dict) and isinstance(high, dict):

            self.defaultshandling("swirr", 0.0, [low, base, high])
            self.defaultshandling("swcr", 0.0, [low, base, high])
            self.defaultshandling("sorg", 0.0, [low, base, high])
            self.defaultshandling("sgcr", 0.0, [low, base, high])
            self.defaultshandling("kroend", 1.0, [low, base, high])
            self.defaultshandling("krwmax", 1.0, [low, base, high])
            self.defaultshandling("krgend", 1.0, [low, base, high])
            self.defaultshandling("krgmax", 1.0, [low, base, high])

            # Initialize saturation ranges for all curves
            self.low = WaterOilGas(
                swirr=low["swirr"],
                swl=low["swl"],
                sorw=low["sorw"],
                sorg=low["sorg"],
                sgcr=low["sgcr"],
                swcr=low["swcr"],
                h=h,
                tag=tag,
            )
            self.base = WaterOilGas(
                swirr=base["swirr"],
                swl=base["swl"],
                sorw=base["sorw"],
                sorg=base["sorg"],
                sgcr=base["sgcr"],
                swcr=base["swcr"],
                h=h,
                tag=tag,
            )
            self.high = WaterOilGas(
                swirr=high["swirr"],
                swl=high["swl"],
                sorw=high["sorw"],
                sorg=high["sorg"],
                sgcr=high["sgcr"],
                swcr=high["swcr"],
                h=h,
                tag=tag,
            )

            # Add water and oil curves
            self.low.wateroil.add_LET_water(
                l=low["Lw"],
                e=low["Ew"],
                t=low["Tw"],
                krwend=low["krwend"],
                krwmax=low["krwmax"],
            )
            self.base.wateroil.add_LET_water(
                l=base["Lw"],
                e=base["Ew"],
                t=base["Tw"],
                krwend=base["krwend"],
                krwmax=base["krwmax"],
            )
            self.high.wateroil.add_LET_water(
                l=high["Lw"],
                e=high["Ew"],
                t=high["Tw"],
                krwend=high["krwend"],
                krwmax=high["krwmax"],
            )

            self.low.wateroil.add_LET_oil(
                l=low["Lo"], e=low["Eo"], t=low["To"], kroend=low["kroend"]
            )
            self.base.wateroil.add_LET_oil(
                l=base["Lo"], e=base["Eo"], t=base["To"], kroend=base["kroend"]
            )
            self.high.wateroil.add_LET_oil(
                l=high["Lo"], e=high["Eo"], t=high["To"], kroend=high["kroend"]
            )

            # Add gas and oil curves:
            self.low.gasoil.add_LET_gas(
                l=low["Lg"],
                e=low["Eg"],
                t=low["Tg"],
                krgend=low["krgend"],
                krgmax=low["krgmax"],
            )
            self.base.gasoil.add_LET_gas(
                l=base["Lg"],
                e=base["Eg"],
                t=base["Tg"],
                krgend=base["krgend"],
                krgmax=base["krgmax"],
            )
            self.high.gasoil.add_LET_gas(
                l=high["Lg"],
                e=high["Eg"],
                t=high["Tg"],
                krgend=high["krgend"],
                krgmax=high["krgmax"],
            )
            self.low.gasoil.add_LET_oil(
                l=low["Log"], e=low["Eog"], t=low["Tog"], kroend=low["kroend"]
            )
            self.base.gasoil.add_LET_oil(
                l=base["Log"], e=base["Eog"], t=base["Tog"], kroend=base["kroend"]
            )
            self.high.gasoil.add_LET_oil(
                l=high["Log"], e=high["Eog"], t=high["Tog"], kroend=high["kroend"]
            )
        else:
            if (
                isinstance(low, WaterOilGas)
                and isinstance(base, WaterOilGas)
                and isinstance(high, WaterOilGas)
            ):

                self.low = low
                self.base = base
                self.high = high

    # User should add capillary pressure explicitly by calling add**
    # on the class objects, or run the following method to add the
    # same to all curves:
    def add_simple_J(self, a=5, b=-1.5, poro_ref=0.25, perm_ref=100, drho=300, g=9.81):
        """Add (identical) simplified J-function to all water-oil
        curves in the SCAL recommendation set"""
        self.low.wateroil.add_simple_J(
            a=a, b=b, poro_ref=poro_ref, perm_ref=perm_ref, drho=drho, g=g
        )
        self.base.wateroil.add_simple_J(
            a=a, b=b, poro_ref=poro_ref, perm_ref=perm_ref, drho=drho, g=g
        )
        self.high.wateroil.add_simple_J(
            a=a, b=b, poro_ref=poro_ref, perm_ref=perm_ref, drho=drho, g=g
        )

    def interpolate(self, parameter, parameter2=None, h=0.02):
        """Interpolate between low, base and high
        parameter = -1 reproduces low curve
        parameter = 0 reproduces base curve
        parameter = 1 reproduces high curve

        Interpolation is performed pointwise in the "relperm" direction
        for each saturation point. During interpolation, endpoints are
        hard to handle, and only swirr is attempted preserved.

        Interpolation is linear in relperm-direction, and will thus not be
        linear in log-relperm-direction

        This method returns an WaterOilGasTable object which can be
        realized into printed tables. No attempt is made to
        parametrize the interpolant in L,E,T parameter space.

        If a second parameter is supplied ("parameter2") this is used
        for the gas-oil interpolation. This enables the gas-oil
        interpolant to be fully uncorrelated to the water-oil
        interpolant. CHECK how this affects endpoints and Eclipse
        consistency!!

        """

        if parameter2 is not None:
            gasparameter = parameter2
        else:
            gasparameter = parameter

        # Initialize wateroil and gasoil curves to be filled with
        # interpolated curves:
        interpolant = WaterOilGas()

        if abs(parameter) > 1.0:
            logging.error("Interpolation parameter must be in [-1,1]")
            interpolant.wateroil = None
            raise AssertionError
        elif np.isclose(parameter, 0.0):
            interpolant.wateroil = self.base.wateroil
        elif np.isclose(parameter, -1.0):
            interpolant.wateroil = self.low.wateroil
        elif np.isclose(parameter, 1.0):
            interpolant.wateroil = self.high.wateroil
        elif parameter < 0.0:
            curve1 = copy.deepcopy(self.base.wateroil)
            curve2 = copy.deepcopy(self.low.wateroil)
            param_transf = -parameter  # 0 gives base, 1 gives low.
            swl = (
                curve1.table["sw"][0] * (1 - param_transf)
                + curve2.table["sw"][0] * param_transf
            )
            interpolant.wateroil = WaterOil(
                swirr=swl,
                swl=swl,
                sorw=0.0,
                h=h,
                tag=self.tag + " interpolant at %g" % parameter,
            )
            self.__interpolator(
                interpolant.wateroil,
                curve1,
                curve2,
                param_transf,
                "sw",
                "krw",
                "krow",
                "pc",
            )
        elif parameter > 0.0:
            curve1 = copy.deepcopy(self.base.wateroil)
            curve2 = copy.deepcopy(self.high.wateroil)
            param_transf = parameter  # 0 gives base, 1 gives high
            swl = (
                curve1.table["sw"][0] * (1 - param_transf)
                + curve2.table["sw"][0] * param_transf
            )
            interpolant.wateroil = WaterOil(
                swirr=swl,
                swl=swl,
                sorw=0.0,
                h=h,
                tag=self.tag + " interpolant at %g" % parameter,
            )
            self.__interpolator(
                interpolant.wateroil,
                curve1,
                curve2,
                param_transf,
                "sw",
                "krw",
                "krow",
                "pc",
            )

        # Gas-oil interpolation
        # We need swl from the interpolated WaterOil object.
        swl = interpolant.wateroil.swl
        if abs(gasparameter) > 1.0:
            logging.error("Interpolation parameter must be in [-1,1]")
            interpolant.gasoil = None
            raise AssertionError
        elif np.isclose(gasparameter, 0.0):
            interpolant.gasoil = self.base.gasoil
        elif np.isclose(gasparameter, -1.0):
            interpolant.gasoil = self.low.gasoil
        elif np.isclose(gasparameter, 1.0):
            interpolant.gasoil = self.high.gasoil
        elif gasparameter < 0.0:
            curve1 = copy.deepcopy(self.base.gasoil)
            curve2 = copy.deepcopy(self.low.gasoil)
            gas_param_transf = -1 * gasparameter  # 0 gives base, 1 gives low.
            # We have to use the extreme, not interpolated sgcr.
            sgcr = min(curve1.sgcr, curve2.sgcr)
            interpolant.gasoil = GasOil(
                sgcr=sgcr,
                swl=swl,
                sorg=0.0,
                h=h,
                tag=self.tag + " interpolant at %g" % gasparameter,
            )
            self.__interpolator(
                interpolant.gasoil,
                curve1,
                curve2,
                gas_param_transf,
                "sg",
                "krg",
                "krog",
                "pc",
            )
            interpolant.gasoil.resetsorg()
        elif gasparameter > 0.0:
            curve1 = copy.deepcopy(self.base.gasoil)
            curve2 = copy.deepcopy(self.high.gasoil)
            gas_param_transf = gasparameter  # 0 gives base, 1 gives high
            # We have to use the extreme, not interpolated sgcr.
            sgcr = min(curve1.sgcr, curve2.sgcr)
            interpolant.gasoil = GasOil(
                sgcr=sgcr,
                swl=swl,
                sorg=0.0,
                h=h,
                tag=self.tag + " interpolant at %g" % gasparameter,
            )
            self.__interpolator(
                interpolant.gasoil,
                curve1,
                curve2,
                gas_param_transf,
                "sg",
                "krg",
                "krog",
                "pc",
            )
            interpolant.gasoil.resetsorg()

        return interpolant

    def __interpolator(
        self,
        tableobject,
        curve1,
        curve2,
        parameter,
        sat="sw",
        kr1="krw",
        kr2="krow",
        pc="pc",
    ):
        """Interpolates between two curves using one parameter between 0 and
        1, does not care if it is water-oil or gas-oil.  First
        argument is the wateroil or gasoil object it is to populate.

        0 will return curve1
        1 will return curve2

        No return value, but modifies the object pointed to by first handle

        """

        curve1.table.rename(columns={kr1: kr1 + "_1"}, inplace=True)
        curve2.table.rename(columns={kr1: kr1 + "_2"}, inplace=True)
        curve1.table.rename(columns={kr2: kr2 + "_1"}, inplace=True)
        curve2.table.rename(columns={kr2: kr2 + "_2"}, inplace=True)
        curve1.table.rename(columns={pc: pc + "_1"}, inplace=True)
        curve2.table.rename(columns={pc: pc + "_2"}, inplace=True)

        # Result data container:
        satresult = pd.DataFrame(data=tableobject.table[sat], columns=[sat])

        # Merge swresult with curve1 and curve2, and interpolate all
        # columns in sw:
        intdf = (
            pd.concat([curve1.table, curve2.table, satresult], sort=True)
            .set_index(sat)
            .sort_index()
            .interpolate(method="slinear")
            .fillna(method="bfill")
            .fillna(method="ffill")
        )

        # Normalized saturations does not make sense for the
        # interpolant, remove:
        for col in ["swn", "son", "swnpc", "H", "J"]:
            if col in intdf.columns:
                del intdf[col]

        intdf[kr1] = intdf[kr1 + "_1"] * (1 - parameter) + intdf[kr1 + "_2"] * parameter
        intdf[kr2] = intdf[kr2 + "_1"] * (1 - parameter) + intdf[kr2 + "_2"] * parameter
        if pc + "_1" in curve1.table.columns and pc + "_2" in curve2.table.columns:
            intdf[pc] = (
                intdf[pc + "_1"] * (1 - parameter) + intdf[pc + "_2"] * parameter
            )
        else:
            intdf[pc] = 0

        # Slice out the resulting sw values and columns. Slicing on
        # floating point indices is not robust so we need to slice on an
        # integer version of the sw column
        tableobject.table["swint"] = list(
            map(int, list(map(round, tableobject.table[sat] * SWINTEGERS)))
        )
        intdf["swint"] = list(
            map(int, list(map(round, intdf.index.values * SWINTEGERS)))
        )
        intdf = intdf.reset_index()
        intdf.drop_duplicates("swint", inplace=True)
        intdf.set_index("swint", inplace=True)
        intdf = intdf.loc[tableobject.table["swint"].values]
        intdf = intdf[[sat, kr1, kr2, pc]].reset_index()

        # intdf['swint'] = (intdf['sw'] * SWINTEGERS).astype(int)
        # intdf.drop_duplicates('swint', inplace=True)

        # Populate the WaterOil object
        tableobject.table[kr1] = intdf[kr1]
        tableobject.table[kr2] = intdf[kr2]
        tableobject.table[pc] = intdf[pc]
        tableobject.table.fillna(method="ffill", inplace=True)
        return



# coding: utf-8

from setuptools import setup
import versioneer

from os import path

# Read the contents of README.md, for PyPI
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md")) as f_handle:
    long_description = f_handle.read()

setup(
    name="pyscal",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Generate relative permeability include files for Eclipse reservoir simulator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/equinor/pyscal",
    author="Håvard Berland",
    author_email="havb@equinor.com",
    license="LGPLv3",
    packages=["pyscal"],
    zip_safe=False,
    entry_points={"console_scripts": []},
)



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import
from __future__ import print_function

import math
import logging
import numpy as np
import pandas as pd

from pyscal.constants import EPSILON as epsilon
from pyscal.constants import SWINTEGERS
from pyscal.constants import MAX_EXPONENT


class WaterOil(object):
    """A representation of two-phase properties for oil-water.

    Can hold relative permeability data, and capillary pressure.

    Parametrizations for relative permeability:
     * Corey
     * LET

    For capillary pressure:
     * Simplified J-function

    For object initialization, only saturation endpoints must be inputted,
    and saturation resolution. An optional string can be added as a 'tag'
    that can be used when outputting.

    Relative permeability and/or capillary pressure can be added through
    parametrizations, or from a dataframe (will incur interpolation).

    Can be dumped as include files for Eclipse/OPM and Nexus simulators.
    """

    def __init__(self, swirr=0.0, swl=0.0, swcr=0.0, sorw=0.0, h=0.01, tag=""):
        """Sets up the saturation range. Swirr is only relevant
        for the capillary pressure, not for relperm data."""

        assert -epsilon < swirr < 1.0 + epsilon
        assert -epsilon < swl < 1.0 + epsilon
        assert -epsilon < swcr < 1.0 + epsilon
        assert -epsilon < sorw < 1.0 + epsilon
        assert epsilon < h <= 1
        assert swl < 1 - sorw
        assert swcr < 1 - sorw
        assert swirr < 1 - sorw
        if not isinstance(tag, str):
            tag = ""
        self.swirr = swirr
        self.swl = max(swl, swirr)  # Cannot allow swl < swirr. Warn?
        if not np.isclose(sorw, 0) and sorw < 1 / SWINTEGERS:
            # Give up handling sorw very close to zero
            sorw = 0.0
        if self.swl < swcr < self.swl + 1 / SWINTEGERS + epsilon:
            # Give up handling swcr so close to swl
            swcr = self.swl
        self.swcr = max(self.swl, swcr)  # Cannot allow swcr < swl. Warn?
        self.sorw = sorw
        self.h = h
        self.tag = tag
        sw = list(np.arange(self.swl, 1 - sorw, h)) + [self.swcr] + [1 - sorw] + [1]
        self.table = pd.DataFrame(sw, columns=["sw"])
        # Ensure that we do not have sw values that are too close
        # to each other, determined rougly by the distance 1/10000
        self.table["swint"] = list(
            map(int, list(map(round, self.table["sw"] * SWINTEGERS)))
        )
        self.table.drop_duplicates("swint", inplace=True)
        # Now, sw=1-sorw might be accidentaly dropped, so make sure we
        # have it by replacing the closest value by 1-sorw exactly
        sorwindex = (self.table["sw"] - (1 - self.sorw)).abs().sort_values().index[0]
        self.table.loc[sorwindex, "sw"] = 1 - self.sorw

        # Same for sw=swcr:
        swcrindex = (self.table["sw"] - (self.swcr)).abs().sort_values().index[0]
        self.table.loc[swcrindex, "sw"] = self.swcr

        # If sw=1 was dropped, then sorw was close to zero:
        if not np.isclose(self.table["sw"].max(), 1.0):
            # Add it as an extra row:
            self.table.loc[len(self.table) + 1, "sw"] = 1.0

        self.table.sort_values(by="sw", inplace=True)
        self.table.reset_index(inplace=True)
        self.table = self.table[["sw"]]  # Drop the swint column

        # Normalize for krw:
        self.table["swn"] = (self.table.sw - self.swcr) / (1 - self.swcr - self.sorw)
        # Normalize for krow:
        self.table["son"] = (1 - self.table.sw - self.sorw) / (
            1 - self.sorw - self.swcr
        )

        # Different normalization for Sw used for capillary pressure
        self.table["swnpc"] = (self.table.sw - swirr) / (1 - swirr)

        self.swcomment = "-- swirr=%g swl=%g swcr=%g sorw=%g\n" % (
            self.swirr,
            self.swl,
            self.swcr,
            self.sorw,
        )
        self.krwcomment = ""
        self.krowcomment = ""
        self.pccomment = ""

        logging.info(
            "Initialized WaterOil with %s saturation points", str(len(self.table))
        )

    def add_oilwater_fromtable(
        self,
        df,
        swcolname="Sw",
        krwcolname="krw",
        krowcolname="krow",
        pccolname="pcow",
        krwcomment="",
        krowcomment="",
        pccomment="",
    ):
        """Interpolate relpermdata from a dataframe.

        The saturation range with endpoints must be set up beforehand,
        and must be compatible with the tabular input. The tabular
        input will be interpolated to the initialized Sw-table

        If you have krw and krow in different dataframes, call this
        function twice

        Calling function is responsible for checking if any data was
        actually added to the table.

        The python package ecl2df has a tool for converting Eclipse input
        files to dataframes.

        Args:
            df: Pandas dataframe containing data
            swcolname: string, column name with the saturation data in the dataframe.
            krwcolname: string, name of the column with krw
            krowcolname: string
            pccolname: string
            krwcomment: string
            krowcomment: string
            pccomment: string
        """
        from scipy.interpolate import PchipInterpolator

        if swcolname not in df:
            logging.critical(
                swcolname + " not found in dataframe, can't read table data"
            )
            raise ValueError
        if krwcolname in df:
            pchip = PchipInterpolator(
                df[swcolname].astype(float), df[krwcolname].astype(float)
            )
            self.table["krw"] = pchip(self.table.sw)
            self.krwcomment = "-- krw from tabular input" + krwcomment + "\n"
        if krowcolname in df:
            pchip = PchipInterpolator(
                df[swcolname].astype(float), df[krowcolname].astype(float)
            )
            self.table["krow"] = pchip(self.table.sw)
            self.krowcomment = "-- krow from tabular input" + krowcomment + "\n"
        if pccolname in df:
            pchip = PchipInterpolator(
                df[swcolname].astype(float), df[pccolname].astype(float)
            )
            self.table["pc"] = pchip(self.table.sw)
            self.pccomment = "-- pc from tabular input" + pccomment + "\n"

    def add_corey_water(self, nw=2, krwend=1, krwmax=None):
        """ Add krw data through the Corey parametrization

        A column named 'krw' will be added. If it exists, it will
        be replaced.

        It is assumed that there are no sw points between
        sw=1-sorw and sw=1, which should give linear
        interpolations in simulators. The corey parameter
        applies up to 1-sorw.

        krwmax will be ignored if sorw is close to zero

        Args:
            nw (float): Corey parameter for water.
            krwend (float): value of krw at 1 - sorw.
            krwmax (float): maximal value at Sw=1. Default 1

        """
        assert epsilon < nw < MAX_EXPONENT
        if krwmax:
            assert 0 < krwend <= krwmax <= 1.0
        else:
            assert 0 < krwend <= 1.0

        self.table["krw"] = krwend * self.table.swn ** nw

        self._handle_endpoints_linearpart_water(krwend, krwmax)

        if not krwmax:
            krwmax = 1
        self.krwcomment = "-- Corey krw, nw=%g, krwend=%g, krwmax=%g\n" % (
            nw,
            krwend,
            krwmax,
        )

    def _handle_endpoints_linearpart_water(self, krwend, krwmax=None):
        """Internal utility function to handle krw
        around endpoints.

        """
        # Avoid machine accuracy problems around sorw (monotonicity):
        self.table.loc[self.table["krw"] > krwmax, "krw"] = krwmax

        # We assume there is no points between 1 - sorw and 1, so no
        # need to excplicitly insert linearly interpolated values.
        if self.sorw > 1.0 / SWINTEGERS:
            if not krwmax:
                krwmax = 1.0
            self.table.loc[self.table["sw"] > (1 - self.sorw - epsilon), "krw"] = krwmax
        else:
            if krwmax:
                logging.warning("krwmax ignored when sorw is zero")
            self.table.loc[self.table["sw"] > (1 - self.sorw - epsilon), "krw"] = krwend
        self.table.loc[np.isclose(self.table["sw"], 1 - self.sorw), "krw"] = krwend
        self.table.loc[self.table.sw < self.swcr, "krw"] = 0

    def _handle_endpoints_linearpart_oil(self, kroend, kromax=None):
        """Internal utility function to handle krow
        around endpoints.

        Ensures we obey the endpoints, and linearity where needed"""

        # Linear curve between swl and swcr:
        self.table.loc[self.table["son"] > 1.0 + epsilon, "krow"] = np.nan
        if self.swcr < self.swl + self.h:
            if kromax:
                logging.warning("kromax ignored when swcr is close to swl")
            self.table.loc[self.table["sw"] <= self.swl + epsilon, "krow"] = kroend
        else:
            if not kromax:
                kromax = 1
            self.table.loc[self.table["sw"] <= self.swl + epsilon, "krow"] = kromax
            interp_krow = (
                self.table[["sw", "krow"]]
                .set_index("sw")
                .interpolate(method="index")["krow"]
            )
            self.table.loc[:, "krow"] = interp_krow.values

        # Avoid machine accuracy problems around swcr (monotonicity):
        self.table.loc[self.table["krow"] > kromax, "krow"] = kromax

    def add_LET_water(self, l=2, e=2, t=2, krwend=1, krwmax=None):
        """Add krw data through LET parametrization

        It is assumed that there are no sw points between
        sw=1-sorw and sw=1, which should give linear
        interpolations in simulators. The LET parameters
        apply up to 1-sorw.

        krwmax will be ignored if sorw is close to zero.

        Args:
            l (float): LET parameter
            e (float): LET parameter
            t (float): LET parameter
            krwend (float): value of krw at 1 - sorw
            krwmax (float): maximal value at Sw=1. Default 1
        """
        assert epsilon < l < MAX_EXPONENT
        assert epsilon < e < MAX_EXPONENT
        assert epsilon < t < MAX_EXPONENT
        if krwmax:
            assert 0 < krwend <= krwmax <= 1.0
        else:
            assert 0 < krwend <= 1.0

        self.table["krw"] = (
            krwend
            * self.table.swn ** l
            / ((self.table.swn ** l) + e * (1 - self.table.swn) ** t)
        )
        self._handle_endpoints_linearpart_water(krwend, krwmax)

        if not krwmax:
            krwmax = 1
        self.krwcomment = "-- LET krw, l=%g, e=%g, t=%g, krwend=%g, krwmax=%g\n" % (
            l,
            e,
            t,
            krwend,
            krwmax,
        )

    def add_LET_oil(self, l=2, e=2, t=2, kroend=1, kromax=None):
        """
        Add kro data through LET parametrization

        kromax will be ignored if swcr is close to swl.

        Args:
            l (float): LET parameter
            e (float): LET parameter
            t (float): LET parameter
            kroend (float): value of kro at swcr
            kromax (float): maximal value of kro at sw=swl. Default 1

        Returns:
            None (modifies object)
        """
        assert epsilon < l < MAX_EXPONENT
        assert epsilon < e < MAX_EXPONENT
        assert epsilon < t < MAX_EXPONENT
        if kromax:
            assert 0 < kroend <= kromax <= 1.0
        else:
            assert 0 < kroend <= 1.0

        self.table["krow"] = (
            kroend
            * self.table.son ** l
            / ((self.table.son ** l) + e * (1 - self.table.son) ** t)
        )
        self.table.loc[self.table.sw >= (1 - self.sorw), "krow"] = 0

        self._handle_endpoints_linearpart_oil(kroend, kromax)

        if not kromax:
            kromax = 1
        self.krowcomment = "-- LET krow, l=%g, e=%g, t=%g, kroend=%g, kromax=%g\n" % (
            l,
            e,
            t,
            kroend,
            kromax,
        )

    def add_corey_oil(self, now=2, kroend=1, kromax=None):
        """Add kro data through the Corey parametrization

        Corey applies to the interval between swcr and 1 - sorw

        Curve is linear between swl and swcr, zero above 1 - sorw.

        kromax will be ignored if swcr is close to swl.

        Args:
            now (float): Corey exponent
            kroend (float): kro value at swcr
            kromax (float): kro value at swl
        Returns:
            None (modifies object)
        """
        assert epsilon < now < MAX_EXPONENT
        if kromax:
            assert 0 < kroend <= kromax <= 1.0
        else:
            assert 0 < kroend <= 1.0

        self.table["krow"] = kroend * self.table.son ** now
        self.table.loc[self.table.sw >= (1 - self.sorw), "krow"] = 0

        self._handle_endpoints_linearpart_oil(kroend, kromax)

        if not kromax:
            kromax = 1
        self.krowcomment = "-- Corey krow, now=%g, kroend=%g, kromax=%g\n" % (
            now,
            kroend,
            kromax,
        )

    def add_simple_J(self, a=5, b=-1.5, poro_ref=0.25, perm_ref=100, drho=300, g=9.81):
        """Add capillary pressure function from a simplified J-function

        This is the 'inverse' or 'RMS' version of the a and b, the formula
        is

            J = a S_w^b

        J is not dimensionless.
        Doc: https://wiki.equinor.com/wiki/index.php/Res:Water_saturation_from_Leverett_J-function

        poro_ref is a fraction, between 0 and 1
        perm_ref is in milliDarcy
        drho has SI units kg/m³. Default value is 300
        g has SI units m/s², default value is 9.81
        """
        assert g >= 0
        assert b < MAX_EXPONENT
        assert b > -MAX_EXPONENT
        assert 0.0 <= poro_ref <= 1.0
        assert perm_ref > 0.0

        if self.swl < epsilon:
            logging.error(
                "swl must larger than zero to avoid infinite capillary pressure"
            )
            raise ValueError

        if b > 0:
            logging.warning(
                "positive b will give increasing capillary pressure with saturation"
            )

        # drho = rwo_w - rho_o, in units g/cc

        # swnpc is a normalized saturation, but normalized with
        # respect to swirr, not to swl (the swirr here is sometimes
        # called 'swirra' - asymptotic swirr)
        self.table["J"] = a * self.table.swnpc ** b
        self.table["H"] = self.table.J * math.sqrt(poro_ref / perm_ref)
        # Scale drho and g from SI units to g/cc and m/s²100
        self.table["pc"] = self.table.H * drho / 1000 * g / 100.0
        self.pccomment = (
            "-- Simplified J function for Pc; \n--   "
            + "a=%g, b=%g, poro_ref=%g, perm_ref=%g mD, drho=%g kg/m³, g=%g m/s²\n"
            % (a, b, poro_ref, perm_ref, drho, g)
        )

    def add_normalized_J(self, a, b, poro, perm, sigma_costau):
        """
        Add capillary pressure in bar through a normalized J-function.

        .. math::

            p_c = \\frac{\left(\\frac{S_w}{a}\\right)^{\\frac{1}{b}}
            \sigma \cos \\tau}{\sqrt{\\frac{k}{\phi}}}

        The Sw saturation used in the formula is normalized with respect
        to the swirr parameter.

        Args:
            a (float): a parameter
            b (float): b exponent (typically negative)
            poro (float): Porosity value, fraction between 0 and 1
            perm (float): Permeability value in mD
            sigma_costau (float): Interfacial tension in mN/m (typical value 30 mN/m)

        Returns:
            None. Modifies pc column in self.table, using bar as pressure unit.
        """
        assert epsilon < abs(a) < MAX_EXPONENT
        assert epsilon < abs(b) < MAX_EXPONENT
        assert epsilon < poro <= 1.0
        assert epsilon < perm
        assert isinstance(sigma_costau, (int, float))

        if b < 0 and np.isclose(self.swirr, self.swl):
            logging.error("swl must be set larger than swirr to avoid infinite p_c")
            raise ValueError("swl must be larger than swirr")

        if abs(b) < 0.01:
            logging.warning(
                "b exponent is very small, risk of infinite capillary pressure"
            )

        if abs(a) < 0.01:
            logging.warning(
                "a parameter is very small, risk of infinite capillary pressure"
            )

        if abs(a) > 5:
            logging.warning(
                "a parameter is very high, risk of infinite capillary pressure"
            )

        # 1 atm is equivalent to 101325 pascal = 1.01325 bar
        # pascal_to_atm = 1.0 / 101325.0  # = 9.86923267e-6
        pascal_to_bar = 1e-5

        perm_D = perm / 1000
        perm_sq_meters = perm_D * 9.869233e-13
        tmp = (self.table["swnpc"] / a) ** (1.0 / b)
        tmp = tmp / math.sqrt(perm_sq_meters / poro)
        tmp = tmp * sigma_costau / 1000  # Converting mN/m to N/m
        self.table["pc"] = tmp * pascal_to_bar
        self.pccomment = (
            "-- Capillary pressure from normalized J-function, in bar\n"
            + "-- a=%g, b=%g, poro=%g, perm=%g mD, sigma_costau=%g mN/m\n"
            % (a, b, poro, perm, sigma_costau)
        )

    def add_skjaeveland_pc(self, cw, co, aw, ao, swr=None, sor=None):
        """Add capillary pressure from the Skjæveland correlation,

        Doc: https://wiki.equinor.com/wiki/index.php/Res:The_Skjaeveland_correlation_for_capillary_pressure

        The implementation is unit independent, units are contained in the
        input constants.

        If swr and sor are not provided, it will be taken from the
        swirr and sorw. Only use different values here if you know
        what you are doing.

        Modifies or adds self.table.pc if succesful.
        Returns false if error occured.

        """
        inputerror = False
        if cw < 0:
            logging.error("cw must be larger or equal to zero")
            inputerror = True
        if co > 0:
            logging.error("co must be less than zero")
            inputerror = True
        if aw <= 0:
            logging.error("aw must be larger than zero")
            inputerror = True
        if ao <= 0:
            logging.error("ao must be larger than zero")
            inputerror = True

        if swr == None:
            swr = self.swirr
        if sor == None:
            sor = self.sorw

        if swr >= 1 - sor:
            logging.error("swr (swirr) must be less than 1 - sor")
            inputerror = True
        if swr < 0 or swr > 1:
            logging.error("swr must be contained in [0,1]")
            inputerror = True
        if sor < 0 or sor > 1:
            logging.error("sor must be contained in [0,1]")
            inputerror = True
        if inputerror:
            return False
        self.pccomment = (
            "-- Skjæveland correlation for Pc;\n"
            + "-- cw=%g, co=%g, aw=%g, ao=%g, swr=%g, sor=%g\n"
            % (cw, co, aw, ao, swr, sor)
        )

        # swnpc is a normalized saturation, but normalized with
        # respect to swirr, not to swl (the swirr here is sometimes
        # called 'swirra' - asymptotic swirr)

        # swnpc is generated upon object initialization, but overwritten
        # here to most likely the same values.
        self.table["swnpc"] = (self.table.sw - swr) / (1 - swr)

        # sonpc is almost like 'son', but swl is not used here:
        self.table["sonpc"] = (1 - self.table.sw - sor) / (1 - sor)

        # The Skjæveland correlation
        self.table.loc[self.table.sw < 1 - sor, "pc"] = cw / (
            self.table.swnpc ** aw
        ) + co / (self.table.sonpc ** ao)

        # From 1-sor, the pc is not defined. We want to extrapolate constantly,
        # but with a twist as Eclipse does not non-monotone capillary pressure:
        self.table["pc"].fillna(value=self.table.pc.min(), inplace=True)
        nanrows = self.table.sw > 1 - sor - epsilon
        self.table.loc[nanrows, "pc"] = (
            self.table.loc[nanrows, "pc"] - self.table.loc[nanrows, "sw"]
        )  # Just deduct sw to make it monotone..

    def add_LET_pc_pd(self, Lp, Ep, Tp, Lt, Et, Tt, Pcmax, Pct):
        """Add a primary drainage LET capillary pressure curve.

        Docs: https://wiki.equinor.com/wiki/index.php/Res:The_LET_correlation_for_capillary_pressure

        Note that Pc where Sw > 1 - sorw will appear linear because
        there are no saturation points in that interval.
        """

        # The "forced part"
        self.table["Ffpcow"] = (1 - self.table.swnpc) ** Lp / (
            (1 - self.table.swnpc) ** Lp + Ep * self.table.swnpc ** Tp
        )

        # The gradual rise part:
        self.table["Ftpcow"] = self.table.swnpc ** Lt / (
            self.table.swnpc ** Lt + Et * (1 - self.table.swnpc) ** Tt
        )

        # Putting it together:
        self.table["pc"] = (
            (Pcmax - Pct) * self.table.Ffpcow - Pct * self.table.Ftpcow + Pct
        )

        # Special handling of the interval [0,swirr]
        self.table.loc[self.table.swn < epsilon, "pc"] = Pcmax
        self.pccomment = (
            "-- LET correlation for primary drainage Pc;\n"
            + "-- Lp=%g, Ep=%g, Tp=%g, Lt=%g, Et=%g, Tt=%g, Pcmax=%g, Pct=%g\n"
            % (Lp, Ep, Tp, Lt, Et, Tt, Pcmax, Pct)
        )

    def add_LET_pc_imb(self, Ls, Es, Ts, Lf, Ef, Tf, Pcmax, Pcmin, Pct):
        """Add an imbition LET capillary pressure curve.

        Docs: https://wiki.equinor.com/wiki/index.php/Res:The_LET_correlation_for_capillary_pressure
        """

        # Normalized water saturation including sorw
        self.table["swnpco"] = (self.table.sw - self.swirr) / (
            1 - self.sorw - self.swirr
        )

        # The "forced part"
        self.table["Fficow"] = (1 - self.table.swnpco) ** Ls / (
            (1 - self.table.swnpco) ** Ls + Es * self.table.swnpco ** Ts
        )

        # The gradual rise part:
        self.table["Fticow"] = self.table.swnpco ** Lf / (
            self.table.swnpco ** Lf + Ef * (1 - self.table.swnpco) ** Tf
        )

        # Putting it together:
        self.table["pc"] = (
            (Pcmax + Pct) * self.table.Fficow - (Pcmin - Pct) * self.table.Fticow - Pct
        )

        # Special handling of the interval [0,swirr]
        self.table.loc[self.table.swnpco < epsilon, "pc"] = Pcmax
        # and [1-sorw,1]
        self.table.loc[self.table.swnpco > 1 - epsilon, "pc"] = self.table.pc.min()
        self.pccomment = (
            "-- LET correlation for imbibition Pc;\n"
            + "-- Ls=%g, Es=%g, Ts=%g, Lf=%g, Ef=%g, Tf=%g, Pcmax=%g, Pcmin=%g, Pct=%g\n"
            % (Ls, Es, Ts, Lf, Ef, Tf, Pcmax, Pcmin, Pct)
        )

    def selfcheck(self):
        """Check validities of the data in the table.

        If you call SWOF, this function must not return False
        """
        error = False
        if not (self.table.sw.diff().dropna().round(10) > -epsilon).all():
            logging.error("sw data not strictly increasing")
            error = True
        if not (self.table.krw.diff().dropna().round(10) >= -epsilon).all():
            logging.error("krw data not monotonically increasing")
            error = True
        if (
            "krow" in self.table.columns
            and not (self.table.krow.diff().dropna().round(10) <= epsilon).all()
        ):
            logging.error("krow data not monotonically decreasing")
            error = True
        if "pc" in self.table.columns and self.table.pc[0] > -epsilon:
            if not (self.table.pc.diff().dropna().round(10) < epsilon).all():
                logging.error("pc data not strictly decreasing")
                error = True
        if "pc" in self.table.columns and np.isinf(self.table.pc.max()):
            logging.error("pc goes to infinity. Maybe swirr=swl?")
            error = True
        for col in list(set(["sw", "krw", "krow"]) & set(self.table.columns)):
            if not (
                (round(min(self.table[col]), 10) >= -epsilon)
                and (round(max(self.table[col]), 10) <= 1 + epsilon)
            ):
                logging.error("%s data should be contained in [0,1]", col)
                error = True
        if error:
            return False
        else:
            logging.info("WaterOil object is checked to be valid")
            return True

    def SWOF(self, header=True, dataincommentrow=True):
        if not self.selfcheck():
            return
        string = ""
        if "pc" not in self.table.columns:
            self.table["pc"] = 0
            self.pccomment = "-- Zero capillary pressure\n"
        if header:
            string += "SWOF\n"
        string += "-- " + self.tag + "\n"
        string += "-- Sw Krw Krow Pc\n"
        if dataincommentrow:
            string += self.swcomment
            string += self.krwcomment
            string += self.krowcomment
            string += "-- krw = krow @ sw=%1.5f\n" % self.crosspoint()
            string += self.pccomment
        string += self.table[["sw", "krw", "krow", "pc"]].to_csv(
            sep=" ", float_format="%1.7f", header=None, index=False
        )
        string += "/\n"  # Empty line at the end
        return string

    def SWFN(self, header=True, dataincommentrow=True):
        if not self.selfcheck():
            return
        string = ""
        if "pc" not in self.table.columns:
            self.table["pc"] = 0
            self.pccomment = "-- Zero capillary pressure\n"
        if header:
            string += "SWFN\n"
        string += "-- " + self.tag + "\n"
        string += "-- Sw Krw Pc\n"
        if dataincommentrow:
            string += self.swcomment
            string += self.krwcomment
            if "krow" in self.table.columns:
                string += "-- krw = krow @ sw=%1.5f\n" % self.crosspoint()
            string += self.pccomment
        string += self.table[["sw", "krw", "pc"]].to_csv(
            sep=" ", float_format="%1.7f", header=None, index=False
        )
        string += "/\n"  # Empty line at the end
        return string

    def WOTABLE(self, header=True, dataincommentrow=True):
        """Return a string for a Nexus WOTABLE"""
        string = ""
        if "pc" not in self.table.columns:
            self.table["pc"] = 0
            self.pccomment = "-- Zero capillary pressure\n"
        if header:
            string += "WOTABLE\n"
            string += "SW KRW KROW PC\n"
        if dataincommentrow:
            string += self.swcomment.replace("--", "!")
            string += self.krwcomment.replace("--", "!")
            string += self.krowcomment.replace("--", "!")
            string += "! krw = krow @ sw=%1.5f\n" % self.crosspoint()
            string += self.pccomment.replace("--", "!")
        string += self.table[["sw", "krw", "krow", "pc"]].to_csv(
            sep=" ", float_format="%1.7f", header=None, index=False
        )
        return string

    def plotpc(
        self,
        ax=None,
        color="blue",
        alpha=1,
        label=None,
        linewidth=1,
        linestyle="-",
        logscale=False,
    ):
        """Plot capillary pressure (pc) a supplied matplotlib axis"""
        import matplotlib.pyplot as plt
        import matplotlib

        if not ax:
            matplotlib.style.use("ggplot")
            _, useax = plt.subplots()
        else:
            useax = ax
        self.table.plot(
            ax=useax,
            x="sw",
            y="pc",
            c=color,
            alpha=alpha,
            legend=None,
            linewidth=linewidth,
            linestyle=linestyle,
        )
        if logscale:
            useax.set_yscale("log")
        if not ax:
            plt.show()

    def plotkrwkrow(
        self, ax=None, color="blue", alpha=1, label=None, linewidth=1, linestyle="-"
    ):
        """Plot krw and krow on a supplied matplotlib axis"""
        import matplotlib.pyplot as plt
        import matplotlib

        if not ax:
            matplotlib.style.use("ggplot")
            _, useax = plt.subplots()
        else:
            useax = ax
        self.table.plot(
            ax=useax,
            x="sw",
            y="krw",
            c=color,
            alpha=alpha,
            legend=None,
            linewidth=linewidth,
            linestyle=linestyle,
        )
        self.table.plot(
            ax=useax,
            x="sw",
            y="krow",
            c=color,
            alpha=alpha,
            legend=None,
            linewidth=linewidth,
            linestyle=linestyle,
        )
        if not ax:
            plt.show()

    def crosspoint(self):
        """Locate and return the saturation point where krw = krow

        Accuracy of this crosspoint depends on the resolution chosen
        when initializing the saturation range
       """

        # Make a copy for calculations
        tmp = pd.DataFrame(self.table[["sw", "krw", "krow"]])
        tmp.loc[:, "krwminuskrow"] = tmp["krw"] - tmp["krow"]

        # Add a zero value for the difference column, and interpolate
        # the sg column to the zero value
        zerodf = pd.DataFrame(index=[len(tmp)], data={"krwminuskrow": 0.0})
        tmp = pd.concat([tmp, zerodf], sort=True)
        # When Pandas is upgraded for all users:
        # tmp = pd.concat([tmp, zerodf], sort=True)
        tmp.set_index("krwminuskrow", inplace=True)
        tmp.interpolate(method="slinear", inplace=True)

        return tmp[np.isclose(tmp.index, 0.0)].sw.values[0]



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import
from __future__ import print_function

import logging
import numpy as np
import pandas as pd

from pyscal.constants import EPSILON as epsilon
from pyscal.constants import SWINTEGERS
from pyscal.constants import MAX_EXPONENT


class GasOil(object):
    """Object to represent two-phase properties for gas and oil.

    Parametrizations available for relative permeability:

     * Corey
     * LET

    or data can alternatively be read in from tabulated data
    (as Pandas DataFrame).

    No support (yet) to add capillary pressure.

    krgend can be anchored both to `1-swl-sorg` and to `1-swl`. Default is
    to anchor to `1-swl-sorg`. If the krgendanchor argument is something
    else than the string `sorg`, it will be anchored to `1-swl`.

    Code duplication warning: Code is analogous to WaterOil, but with
    some subtle details sufficiently different for now warranting its
    own code (no easy inheritance)

    Arguments:
        swirr (float): Absolute minimal water saturation at infinite capillary
            pressure.
            Not in use currently, except for in informational headers and
            for consistency checks.
        swl (float): First water saturation point in water tables. In GasOil, it
            is used to obtain the normalized oil and gas saturation.
        sgcr (float): Critical gas saturation. Gas will not be mobile before the
            gas saturation is above this value.
        sorg (float): Residual oil saturation after gas flooding. At this oil
            saturation, the oil has zero relative permeability.
        krgendanchor (str): Set to `sorg` or something else, where to
            anchor `krgend`. If `sorg`, then the normalized gas
            saturation will be equal to 1 at `1 - swl - sgcr - sorg`,
            if not, it will be 1 at `1 - swl - sgcr`. If `sorg` is zero
            it does not matter.
        h (float): Saturation step-length in the outputted table.
        tag (str): Optional string identifier, only used in comments.
    """

    def __init__(
        self, swirr=0, sgcr=0.0, h=0.01, swl=0.0, sorg=0.0, tag="", krgendanchor="sorg"
    ):
        assert epsilon < h <= 1
        assert -epsilon < swirr < 1.0 + epsilon
        assert -epsilon < sgcr < 1
        assert -epsilon < swl < 1
        assert -epsilon < sorg < 1
        if not isinstance(tag, str):
            tag = ""
        assert isinstance(krgendanchor, str)

        self.h = h
        swl = max(swl, swirr)  # Can't allow swl < swirr, should we warn user?
        self.swl = swl
        self.swirr = swirr
        if not np.isclose(sorg, 0.0) and sorg < 1.0 / SWINTEGERS:
            # Too small but nonzero sorg gives too many numerical issues.
            logging.warning("sorg was close to zero, set to zero")
            sorg = 0.0
        self.sorg = sorg

        self.sgcr = sgcr
        self.tag = tag
        if not 1 - sorg - swl > 0:
            raise Exception(
                "No saturation range left " + "after endpoints, check input"
            )
        if krgendanchor in ["sorg", ""]:
            self.krgendanchor = krgendanchor
        else:
            logging.warning("Unknown krgendanchor %s, ignored", str(krgendanchor))
            self.krgendanchor = ""
        if np.isclose(sorg, 0.0) and self.krgendanchor == "sorg":
            # This is too much info..
            self.krgendanchor = ""  # This is critical to avoid bugs due to numerics.

        sg = (
            [0]
            + [sgcr]
            + list(np.arange(sgcr + h, 1 - swl, h))
            + [1 - sorg - swl]
            + [1 - swl]
        )
        self.table = pd.DataFrame(sg, columns=["sg"])
        self.table["sgint"] = list(
            map(int, list(map(round, self.table["sg"] * SWINTEGERS)))
        )
        self.table.drop_duplicates("sgint", inplace=True)
        # Now sg=1-sorg-swl might be accidentally dropped, so make sure we
        # have it by replacing the closest value by 1 - sorg exactly
        sorgindex = (
            (self.table["sg"] - (1 - self.sorg - self.swl)).abs().sort_values().index[0]
        )
        self.table.loc[sorgindex, "sg"] = 1 - self.sorg - self.swl

        # Same for sg=sgcr
        sgcrindex = (self.table["sg"] - (self.sgcr)).abs().sort_values().index[0]
        self.table.loc[sgcrindex, "sg"] = self.sgcr
        if sgcrindex == 0 and sgcr > 0.0:
            # Need to conserve sg=0
            zero_row = pd.DataFrame({"sg": 0}, index=[0])
            self.table = pd.concat([zero_row, self.table], sort=False).reset_index(
                drop=True
            )

        # If sg=1-swl was dropped, then sorg was close to zero:
        if not np.isclose(self.table["sg"].max(), 1 - self.swl):
            # Add it as an extra row:
            self.table.loc[len(self.table) + 1, "sg"] = 1 - self.swl

        self.table.sort_values(by="sg", inplace=True)
        self.table.reset_index(inplace=True)
        self.table = self.table[["sg"]]
        self.table["sl"] = 1 - self.table["sg"]
        if krgendanchor == "sorg":
            # Normalized sg (sgn) is 0 at sgcr, and 1 at 1-swl-sorg
            self.table["sgn"] = (self.table["sg"] - sgcr) / (1 - swl - sgcr - sorg)
        else:
            self.table["sgn"] = (self.table["sg"] - sgcr) / (1 - swl - sgcr)

        # Normalized oil saturation should be 0 at 1-sorg, and 1 at swl+sgcr
        self.table["son"] = (self.table["sl"] - sorg - swl) / (1 - sorg - swl - sgcr)
        self.sgcomment = (
            '-- swirr=%g, sgcr=%g, swl=%g, sorg=%g, krgendanchor="%s"\n'
            % (self.swirr, self.sgcr, self.swl, self.sorg, self.krgendanchor)
        )
        self.krgcomment = ""
        self.krogcomment = ""
        self.pccomment = ""

        logging.info(
            "Initialized GasOil with %s saturation points", str(len(self.table))
        )

    def resetsorg(self):
        """Recalculate sorg in case it has table data has been manipulated"""
        if "krog" in self.table.columns:
            self.sorg = (
                1 - self.swl - self.table[np.isclose(self.table.krog, 0.0)].min()["sg"]
            )
            self.sgcomment = (
                '-- swirr=%g, sgcr=%g, swl=%g, sorg=%g\n, krgendanchor="%s"'
                % (self.swirr, self.sgcr, self.swl, self.sorg, self.krgendanchor)
            )

    def add_gasoil_fromtable(
        self,
        df,
        sgcolname="Sg",
        krgcolname="krg",
        krogcolname="krog",
        pccolname="pcog",
        krgcomment="",
        krogcomment="",
        pccomment="",
    ):
        """Interpolate relpermdata from a dataframe.

        The saturation range with endpoints must be set up beforehand,
        and must be compatible with the tabular input. The tabular
        input will be interpolated to the initialized Sg-table.
        IMPORTANT: Set sgcr and swl to sensible values.

        If you have krg and krog in different dataframes, call this
        function twice

        Calling function is responsible for checking if any data was
        actually added to the table.

        The dataframe input can be constructed using e.g. swof2csv functionality

        """
        from scipy.interpolate import PchipInterpolator

        if sgcolname not in df:
            logging.critical(
                sgcolname + " not found in dataframe, " + "can't read table data"
            )
            raise ValueError
        swlfrominput = 1 - df[sgcolname].max()
        if abs(swlfrominput - self.swl) < epsilon:
            logging.warning(
                "swl and 1-max(sg) from incoming table does not seem compatible"
            )
            logging.warning("         Do not trust the result near the endpoint.")
        if krgcolname in df:
            pchip = PchipInterpolator(
                df[sgcolname].astype(float), df[krgcolname].astype(float)
            )
            # Do not extrapolate this data. We will bfill and ffill afterwards
            self.table["krg"] = pchip(self.table.sg, extrapolate=False)
            self.table["krg"].fillna(method="ffill", inplace=True)
            self.table["krg"].fillna(method="bfill", inplace=True)
            self.krgcomment = "-- krg from tabular input" + krgcomment + "\n"
        if krogcolname in df:
            pchip = PchipInterpolator(
                df[sgcolname].astype(float), df[krogcolname].astype(float)
            )
            self.table["krog"] = pchip(self.table.sg, extrapolate=False)
            self.table["krog"].fillna(method="ffill", inplace=True)
            self.table["krog"].fillna(method="bfill", inplace=True)
            self.krogcomment = "-- krog from tabular input" + krogcomment + "\n"
        if pccolname in df:
            pchip = PchipInterpolator(
                df[sgcolname].astype(float), df[pccolname].astype(float)
            )
            self.table["pc"] = pchip(self.table.sg, extrapolate=False)
            self.pccomment = "-- pc from tabular input" + pccomment + "\n"

    def _handle_endpoints_linearpart_gas(self, krgend, krgmax=None):
        """Internal utility function to handle krg
        around endpoints.
        """
        self.table.loc[self.table.sg <= self.sgcr, "krg"] = 0

        if self.krgendanchor == "sorg":
            # Linear curve between krgendcanchor and 1-swl if krgend
            # is anchored to sorg
            if not krgmax:
                krgmax = 1
            tmp = pd.DataFrame(self.table[["sg"]])
            tmp["sgendnorm"] = (tmp["sg"] - (1 - (self.sorg + self.swl))) / (self.sorg)
            tmp["krg"] = tmp["sgendnorm"] * krgmax + (1 - tmp["sgendnorm"]) * krgend
            self.table.loc[
                self.table.sg >= (1 - (self.sorg + self.swl + epsilon)), "krg"
            ] = tmp.loc[tmp.sg >= (1 - (self.sorg + self.swl + epsilon)), "krg"]
        else:
            self.table.loc[self.table.sg > (1 - (self.swl + epsilon)), "krg"] = krgend
            if krgmax:
                logging.warning("krgmax ignored when not anchoring to sorg")

    def _handle_endpoints_linearpart_oil(self, kroend, kromax):
        """Internal utility function to handle kro
        around endpoints.
        """

        # Special handling of the part close to sg=1, set to zero.
        self.table.loc[
            self.table["sg"] > 1 - self.sorg - self.swl - epsilon, "krog"
        ] = 0

        # Set kromax at sg=0, but only if sgcr is sufficiently larger than swl.
        if self.sgcr > self.swl + 1.0 / SWINTEGERS:
            if not kromax:
                kromax = 1
            self.table.loc[self.table["sg"] < epsilon, "krog"] = kromax
        else:
            if kromax:
                logging.warning("kromax ignored when sgcr is close to swl")
            self.table.loc[self.table["sg"] < epsilon, "krog"] = kroend

    def add_corey_gas(self, ng=2, krgend=1, krgmax=None):
        """ Add krg data through the Corey parametrization

        A column called 'krg' will be added. If it exists, it will
        be replaced.

        If krgendanchor is set to `sorg` (default), then the normalized
        gas saturation `sgn` (which is what is raised to the power of `ng`)
        is 1 at `1 - swl - sgcr - sorg`. If not, it is 1 at `1 - swl - sgcr`.

        krgmax is only relevant if krgendanchor is 'sorg'
        """
        assert epsilon < ng < MAX_EXPONENT
        assert 0 < krgend <= 1.0
        if krgmax is not None:
            assert 0 < krgend <= krgmax <= 1.0
        self.table["krg"] = krgend * self.table.sgn ** ng

        self._handle_endpoints_linearpart_gas(krgend, krgmax)

        if not krgmax:
            krgmax = 1
        self.krgcomment = "-- Corey krg, ng=%g, krgend=%g, krgmax=%g\n" % (
            ng,
            krgend,
            krgmax,
        )

    def add_corey_oil(self, nog=2, kroend=1, kromax=None):
        """
        Add kro data through the Corey parametrization

        A column named 'kro' will be added to the internal DataFrame,
        replaced if it exists.

        All values above 1 - sorg - swl are set to zero.

        kromax is ignored if sgcr is close to zero

        Arguments:
            nog (float): Corey exponent for oil
            kroend (float): Value for krog at normalized oil saturation 1
            kromax (float): Value for krog at gas saturation 0.

        Returns:
            None (modifies internal class state)
        """
        assert epsilon < nog < MAX_EXPONENT
        if kromax:
            assert 0 < kroend <= kromax <= 1.0
        else:
            assert 0 < kroend <= 1.0

        self.table["krog"] = kroend * self.table.son ** nog

        self._handle_endpoints_linearpart_oil(kroend, kromax)

        if not kromax:
            kromax = 1
        self.krogcomment = "-- Corey krog, nog=%g, kroend=%g, kromax=%g\n" % (
            nog,
            kroend,
            kromax,
        )

    def add_LET_gas(self, l=2, e=2, t=2, krgend=1, krgmax=None):
        """
        Add gas relative permability data through the LET parametrization

        A column called 'krg' will be added, replaced if it does not exist

        If krgendanchor is set to `sorg` (default), then the normalized
        gas saturation `sgn` (which is what is raised to the power of `ng`)
        is 1 at `1 - swl - sgcr - sorg`. If not, it is 1 at `1 - swl - sgcr`

        Arguments:
            l (float): L parameter in LET
            e (float): E parameter in LET
            t (float): T parameter in LET
            krgend (float): Value of krg at normalized gas saturation 1
            krgmax (float): Value of krg at gas saturation 1

        Returns:
            None (modifies internal state)
        """
        assert epsilon < l < MAX_EXPONENT
        assert epsilon < e < MAX_EXPONENT
        assert epsilon < t < MAX_EXPONENT
        if krgmax:
            assert 0 < krgend <= krgmax <= 1.0
        else:
            assert 0 < krgend <= 1.0

        self.table["krg"] = (
            krgend
            * self.table.sgn ** l
            / ((self.table.sgn ** l) + e * (1 - self.table.sgn) ** t)
        )

        self._handle_endpoints_linearpart_gas(krgend, krgmax)

        if not krgmax:
            krgmax = 1
        self.krgcomment = "-- LET krg, l=%g, e=%g, t=%g, krgend=%g, krgmax=%g\n" % (
            l,
            e,
            t,
            krgend,
            krgmax,
        )

    def add_LET_oil(self, l=2, e=2, t=2, kroend=1, kromax=None):
        """Add oil (vs gas) relative permeability data through the Corey
        parametrization.

        A column named 'krog' will be added, replaced if it exists.

        All values where sg > 1 - sorg - swl are set to zero.

        kromax is ignored if sgcr is close to zero

        Arguments:
            l (float): L parameter
            e (float): E parameter
            t (float): T parameter
            kroend (float): The value at gas saturation sgcr
            kromax (float): The value at gas saturation equal to 0.
        """
        assert epsilon < l < MAX_EXPONENT
        assert epsilon < e < MAX_EXPONENT
        assert epsilon < t < MAX_EXPONENT
        if kromax:
            assert 0 < kroend <= kromax <= 1.0
        else:
            assert 0 < kroend <= 1.0

        # LET shape for the interval [sgcr, 1 - swl - sorg]
        self.table["krog"] = (
            kroend
            * self.table["son"] ** l
            / ((self.table["son"] ** l) + e * (1 - self.table["son"]) ** t)
        )

        self._handle_endpoints_linearpart_oil(kroend, kromax)

        if not kromax:
            kromax = 1
        self.krogcomment = "-- LET krog, l=%g, e=%g, t=%g, kroend=%g, kromax=%g\n" % (
            l,
            e,
            t,
            kroend,
            kromax,
        )

    def selfcheck(self):
        """Check validities of the data in the table.

        This is to catch errors that are either physically wrong
        or at least causes Eclipse 100 to stop.

        Returns True if no errors are found, False if at least one
        is found.

        If you call SGOF/SLGOF, this function must not return False.
        """
        error = False
        if not (self.table.sg.diff().dropna() > -epsilon).all():
            logging.error("sg data not strictly increasing")
            error = True
        if not (self.table.krg.diff().dropna() >= -epsilon).all():
            logging.error("krg data not monotonically decreasing")
            error = True

        if (
            "krog" in self.table.columns
            and not (self.table.krog.diff().dropna() <= epsilon).all()
        ):
            logging.error("krog data not monotonically increasing")
            error = True
        if not np.isclose(min(self.table.krg), 0.0):
            logging.error("krg must start at zero")
            error = True
        if "pc" in self.table.columns and self.table.pc[0] > 0:
            if not (self.table.pc.diff().dropna() < epsilon).all():
                logging.error("pc data for gas-oil not strictly deceasing")
                error = True
        if "pc" in self.table.columns and np.isinf(self.table.pc.max()):
            logging.error("pc goes to infinity for gas-oil. ")
            error = True
        for col in list(set(["sg", "krg", "krog"]) & set(self.table.columns)):
            if not (
                (min(self.table[col]) >= -epsilon)
                and (max(self.table[col]) <= 1 + epsilon)
            ):
                logging.error("%s data should be contained in [0,1]", col)
                error = True
        if error:
            return False
        else:
            logging.info("GasOil object is checked to be valid")
            return True

    def SGOF(self, header=True, dataincommentrow=True):
        """
        Produce SGOF input for Eclipse reservoir simulator.

        The columns sg, krg, krog and pc are outputted and
        formatted accordingly.

        Meta-information for the tabulated data are printed
        as Eclipse comments.

        Args:
            header: boolean for whether the SGOF string should be emitted.
                If you have multiple satnums, you should have True only
                for the first (or False for all, and emit the SGOF yourself).
                Defaults to True.
            dataincommentrow: boolean for wheter metadata should be printed,
                defaults to True.
        """
        if not self.selfcheck():
            return
        string = ""
        if "pc" not in self.table.columns:
            self.table["pc"] = 0
            self.pccomment = "-- Zero capillary pressure\n"
        if header:
            string += "SGOF\n"
        string += "-- " + self.tag + "\n"
        string += "-- Sg Krg Krog Pc\n"
        if dataincommentrow:
            string += self.sgcomment
            string += self.krgcomment
            string += self.krogcomment
            string += "-- krg = krog @ sw=%1.5f\n" % self.crosspoint()
            string += self.pccomment
        string += self.table[["sg", "krg", "krog", "pc"]].to_csv(
            sep=" ", float_format="%1.7f", header=None, index=False
        )
        string += "/\n"
        return string

    def slgof_df(self):
        """Slice out an SLGOF table.

        This is a used by the SLGOF() function, it is
        extracted as a single function to facilitate testing."""
        if "pc" not in self.table.columns:
            # Only happens when the SLGOF function is skipped (test code)
            self.table["pc"] = 0
        slgof = (
            self.table[
                self.table["sg"] <= 1 - self.sorg - self.swl + 1.0 / float(SWINTEGERS)
            ]
            .sort_values("sl")[["sl", "krg", "krog", "pc"]]
            .reset_index(drop=True)
        )
        # It is a strict requirement that the first sl value should be swl + sorg,
        # so we modify it if it close. If it is not close, we do not dare to fix
        # it, to ensure we don't cover bugs.
        slgof_sl_mismatch = abs(slgof["sl"].values[0] - (self.sorg + self.swl))
        if slgof_sl_mismatch > epsilon:
            if slgof_sl_mismatch < 2 * 1.0 / float(SWINTEGERS):
                # Repair the table in-place:
                slgof.loc[0, "sl"] = self.sorg + self.swl
                # After modification, we can get duplicate sl values, so drop duplicates:
                slgof["slint"] = list(
                    map(int, list(map(round, slgof["sl"] * SWINTEGERS)))
                )
                slgof.drop_duplicates("slint", inplace=True)
            else:
                # Give up repairing the table:
                logging.critical(
                    "SLGOF does not start at the correct value. Please report as bug."
                )
                logging.error("slgof_sl_mismatch: %f", slgof_sl_mismatcher)
                logging.error(str(slgof.head()))
        return slgof

    def SLGOF(self, header=True, dataincommentrow=True):
        """Produce SLGOF input for Eclipse reservoir simulator.

        The columns sl (liquid saturation), krg, krog and pc are
        outputted and formatted accordingly.

        Meta-information for the tabulated data are printed
        as Eclipse comments.

        Args:
            header: boolean for whether the SLGOF string should be emitted.
                If you have multiple satnums, you should have True only
                for the first (or False for all, and emit the SGOF yourself).
                Defaults to True.
            dataincommentrow: boolean for wheter metadata should be printed,
                defaults to True.
        """
        if not self.selfcheck():
            return
        string = ""
        if "pc" not in self.table.columns:
            self.table["pc"] = 0
            self.pccomment = "-- Zero capillary pressure\n"
        if header:
            string += "SLGOF\n"
        string += "-- " + self.tag + "\n"
        string += "-- Sl Krg Krog Pc\n"
        if dataincommentrow:
            string += self.sgcomment
            string += self.krgcomment
            string += self.krogcomment
            string += "-- krg = krog @ sw=%1.5f\n" % self.crosspoint()
            string += self.pccomment
        string += self.slgof_df().to_csv(
            sep=" ", float_format="%1.7f", header=None, index=False
        )
        string += "/\n"
        return string

    def SGFN(self, header=True, dataincommentrow=True):
        """
        Produce SGFN input for Eclipse reservoir simulator.

        The columns sg, krg, and pc are outputted and
        formatted accordingly.

        Meta-information for the tabulated data are printed
        as Eclipse comments.

        Args:
            header: boolean for whether the SGOF string should be emitted.
                If you have multiple satnums, you should have True only
                for the first (or False for all, and emit the SGOF yourself).
                Defaults to True.
            dataincommentrow: boolean for wheter metadata should be printed,
                defaults to True.
        """
        string = ""
        if "pc" not in self.table.columns:
            self.table["pc"] = 0
            self.pccomment = "-- Zero capillary pressure\n"
        if header:
            string += "SGFN\n"
        string += "-- " + self.tag + "\n"
        string += "-- Sg Krg Pc\n"
        if dataincommentrow:
            string += self.sgcomment
            string += self.krgcomment
            if "krog" in self.table.columns:
                string += "-- krg = krog @ sw=%1.5f\n" % self.crosspoint()
            string += self.pccomment
        string += self.table[["sg", "krg", "pc"]].to_csv(
            sep=" ", float_format="%1.7f", header=None, index=False
        )
        string += "/\n"
        return string

    def GOTABLE(self, header=True, dataincommentrow=True):
        """
        Produce GOTABLE input for the Nexus reservoir simulator.

        The columns sg, krg, krog and pc are outputted and
        formatted accordingly.

        Meta-information for the tabulated data are printed
        as Eclipse comments.

        Args:
            header: boolean for whether the SGOF string should be emitted.
                If you have multiple satnums, you should have True only
                for the first (or False for all, and emit the SGOF yourself).
                Defaults to True.
            dataincommentrow: boolean for wheter metadata should be printed,
                defaults to True.
        """
        string = ""
        if "pc" not in self.table.columns:
            self.table["pc"] = 0
            self.pccomment = "-- Zero capillary pressure\n"
        if header:
            string += "GOTABLE\n"
            string += "SG KRG KROG PC\n"
        if dataincommentrow:
            string += self.sgcomment.replace("--", "!")
            string += self.krgcomment.replace("--", "!")
            string += self.krogcomment.replace("--", "!")
            string += "! krg = krog @ sw=%1.5f\n" % self.crosspoint()
            string += self.pccomment.replace("--", "!")
        string += self.table[["sg", "krg", "krog", "pc"]].to_csv(
            sep=" ", float_format="%1.7f", header=None, index=False
        )
        return string

    def crosspoint(self):
        """Locate and return the saturation point where krg = krog

        Accuracy of this crosspoint depends on the resolution chosen
        when initializing the saturation range (it uses linear
        interpolation to solve for the zero)

        Warning: Code duplication from WaterOil, with
        column names changed only
        """

        # Make a copy for calculations
        tmp = pd.DataFrame(self.table[["sg", "krg", "krog"]])
        tmp.loc[:, "krgminuskrog"] = tmp["krg"] - tmp["krog"]

        # Add a zero value for the difference column, and interpolate
        # the sw column to the zero value
        zerodf = pd.DataFrame(index=[len(tmp)], data={"krgminuskrog": 0.0})
        tmp = pd.concat([tmp, zerodf], sort=True)

        tmp.set_index("krgminuskrog", inplace=True)
        tmp.interpolate(method="slinear", inplace=True)

        return tmp[np.isclose(tmp.index, 0.0)].sg.values[0]

    def plotkrgkrog(
        self, ax=None, color="blue", alpha=1, label=None, linewidth=1, linestyle="-"
    ):
        """Plot krg and krog on a supplied matplotlib axis"""
        import matplotlib.pyplot as plt
        import matplotlib

        if not ax:
            matplotlib.style.use("ggplot")
            _, useax = matplotlib.pyplot.subplots()
        else:
            useax = ax
        self.table.plot(
            ax=useax,
            x="sg",
            y="krg",
            c=color,
            alpha=alpha,
            legend=None,
            linewidth=linewidth,
            linestyle=linestyle,
        )
        self.table.plot(
            ax=useax,
            x="sg",
            y="krog",
            c=color,
            alpha=alpha,
            legend=None,
            linewidth=linewidth,
            linestyle=linestyle,
        )
        if not ax:
            plt.show()
